#
# Copyright (C) Motorola 2002 - All rights reserved
#
# TWiki extension that adds tags for action tracking
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details, published at 
# http://www.gnu.org/copyleft/gpl.html
#
use strict;
use integer;

use CGI;

use TWiki::Func;

use Text::Soundex;
use Time::ParseDate;

use TWiki::Plugins::ActionTrackerPlugin::Attrs;
use TWiki::Plugins::ActionTrackerPlugin::AttrDef;
use TWiki::Plugins::ActionTrackerPlugin::Format;

# Object that represents a single action

# Fields:
# web           Web the action was found in
# topic         Topic the action was found in
# ACTION_NUMBER The number of the action in the topic (deprecated)
# text          The text of the action
# who           The person responsible for the action
# due           When the action is due
# notify        List of people to notify when the action changes
# uid           Unique identifier for the action
# creator       Who created the action
# created       When the action was created
# closer        Who closed the action
# closed        When the action was closed
{ package ActionTrackerPlugin::Action;

  my $now = time();
  my $mainweb = TWiki::Func::getMainWebname();

  # Options for parsedate
  my %pdopt = ( NO_RELATIVE => 1, DATE_REQUIRED => 1, WHOLE => 1 );

  # Types of standard attributes. The 'noload' type tells us
  # not to load the hash from %ACTION attributes, and the 'nomatch' type
  # tells us not to consider it during match operations.
  # Types are defined as a base type and a comma-separated list of
  # format attributes. Two meta-type components 'noload' and 'nomatch'
  # are defined. If an attribute is defined 'noload' no attempt will
  # be made to load a value for it when the action is created. If it
  # is defined 'nomatch' then the attribute will be ignored in match
  # expressions.
  my $dw = 16;
  my $nw = 35;
  my %basetypes =
    (
     changedsince =>
     new ActionTrackerPlugin::AttrDef( 'noload', 0, 0, 0, undef ),
     closed       =>
     new ActionTrackerPlugin::AttrDef( 'date',  $dw, 1, 0, undef ),
     closer       =>
     new ActionTrackerPlugin::AttrDef( 'names', $nw, 1, 0, undef ),
     created      =>
     new ActionTrackerPlugin::AttrDef( 'date',  $dw, 1, 0, undef ),
     creator      =>
     new ActionTrackerPlugin::AttrDef( 'names', $nw, 1, 0, undef ),
     dollar       =>
     new ActionTrackerPlugin::AttrDef( 'noload', 0, 0, 0, undef ),
     due          =>
     new ActionTrackerPlugin::AttrDef( 'date',  $dw, 1, 0, undef ),
     edit         =>
     new ActionTrackerPlugin::AttrDef( 'noload', 0, 0, 0, undef ),
     format       =>
     new ActionTrackerPlugin::AttrDef( 'noload', 0, 0, 0, undef ),
     header       =>
     new ActionTrackerPlugin::AttrDef( 'noload', 0, 0, 0, undef ),
     late         =>
     new ActionTrackerPlugin::AttrDef( 'noload', 0, 0, 0, undef ),
     n            =>
     new ActionTrackerPlugin::AttrDef( 'noload', 0, 0, 0, undef ),
     nop          =>
     new ActionTrackerPlugin::AttrDef( 'noload', 0, 0, 0, undef ),
     notify       =>
     new ActionTrackerPlugin::AttrDef( 'names', $nw, 1, 0, undef ),
     percnt       =>
     new ActionTrackerPlugin::AttrDef( 'noload', 0, 0, 0, undef ),
     quot         =>
     new ActionTrackerPlugin::AttrDef( 'noload', 0, 0, 0, undef ),
     sort         =>
     new ActionTrackerPlugin::AttrDef( 'noload', 0, 0, 0, undef ),
     state        =>
     new ActionTrackerPlugin::AttrDef( 'select', 1, 1, 1,
				       [ "open","closed" ] ),
     text         =>
     new ActionTrackerPlugin::AttrDef( 'noload', 0, 1, 0, undef ),
     topic        =>
     new ActionTrackerPlugin::AttrDef( 'noload', 0, 1, 0, undef ),
     uid          =>
     new ActionTrackerPlugin::AttrDef( 'text',  $nw, 1, 0, undef ),
     web          =>
     new ActionTrackerPlugin::AttrDef( 'noload', 0, 1, 0, undef ),
     who          =>
     new ActionTrackerPlugin::AttrDef( 'names', $nw, 1, 0, undef ),
     within       =>
     new ActionTrackerPlugin::AttrDef( 'noload', 0, 1, 0, undef ),
     ACTION_NUMBER=>
     new ActionTrackerPlugin::AttrDef( 'noload', 0, 0, 0, undef ),
    );

  my %types = %basetypes;

  # PUBLIC Constructor
  sub new {
    my ( $class, $web, $topic, $number, $attrs, $descr ) = @_;
    my $this = {};
    
    my $attr = new ActionTrackerPlugin::Attrs( $attrs );

    # We always have a state, and if it's not defined in the
    # attribute set, and the closed attribute isn't defined,
    # then it takes the value of the first option in the
    # enum for the state attribute. If the closed attribute is
    # defined it takes the last enum.
    $this->{state} = $attr->get( "state" );
    if ( !defined( $this->{state} )) {
      if ( $attr->get( "closed" )) {
	$this->{state} = "closed";
      } else {
	$this->{state} = $types{state}->firstSelect();
      }
    }

    # conditionally load field values, interpreting them
    # according to their type.
    foreach my $key ( keys %$attr ) {
      my $type = getBaseType( $key ) || "noload";
      my $val = $attr->get( $key );
      if ( $type eq "names" && defined( $val )) {
	my @names = split( /[,\s]+/, $val );
	foreach my $n ( @names ) {
	  $n = _canonicalName( $n );
	}
	$this->{$key} = join( ',', @names );
      } elsif ( $type eq "date" ) {
	if ( defined( $val )) {
	  $this->{$key} = Time::ParseDate::parsedate( $val, %pdopt );
	}
      } elsif ( $type !~ "noload" ) {
	# treat as plain string; text, select
	$this->{$key} = $attr->get( $key );
      }
    }

    # do these last so they override and attribute values
    $this->{web} = $web;
    $this->{topic} = $topic;
    $this->{ACTION_NUMBER} = $number;

    $descr =~ s/^\s+//o;
    $descr =~ s/\s+$//o;
    $descr =~ s/\r+//o;

    # Translate newlines in the description to XHTML tags
    $descr =~ s/\n\n/<p \/>/gos;
    $descr =~ s/\n/<br \/>/gos;
    $descr =~ s/%(ACTION\w*)\{/%<nop>$1\{/gso;

    $this->{text} = $descr;

    return bless( $this, $class );
  }

  # PUBLIC STATIC extend the range of types accepted by actions.
  # Return undef if everything went OK, or error message if not.
  # The range of types is extended statically; once extended, there's
  # no way to unextend them.
  # Format of a type def is described in ActionTrackerPlugin.txt
  sub extendTypes {
    my $defs = shift;
    $defs =~ s/^\s*\|//o;
    $defs =~ s/\|\s*$//o;
    foreach my $def ( split( /\s*\|\s*/, $defs )) {
      if ( $def =~ m/^\s*(\w+)\s*,\s*(\w+)\s*(,\s*(\d+)\s*)?(,\s*(.*))?$/o ) {
	my $name = $1;
	my $type = $2;
	my $size = $4;
	my $params = $6;
	my @values;
	my $exists = $types{$name};

	if ( defined( $exists ) && !$exists->isRedefinable() ) {
	  return "Attempt to redefine attribute '$name' in EXTRAS";
	} elsif ( $type eq "select" ) {
	  @values = split( /\s*,\s*/, $params );
	  foreach my $option ( @values ) {
	    $option =~ s/^\"(.*)\"$/$1/o;
	  }
	}
	$types{$name} =
	  new ActionTrackerPlugin::AttrDef( $type, $size, 1, 1, \@values );
      } else {
	return "Bad EXTRAS definition '$def' in EXTRAS";
      }
    }
    return undef;
  }

  # STATIC remove type extensions
  sub unextendTypes {
    %types = %basetypes;
  }

  # PUBLIC get the base type of an attribute name i.e.
  # with the formatting attributes stripped off.
  sub getBaseType {
    my $vbl = shift;
    my $type = $types{$vbl};
    if ( defined( $type ) ) {
      return $type->{type};
    }
    return undef;
  }

  # PUBLIC provided as part of the contract with Format.
  sub getType {
    my ( $this, $name ) = @_;

    return $types{$name};
  }

  # PRIVATE allocate a new UID for this action. This uses a file
  # in the data directory called "atUidReg" that contains the
  # highest UID allocated so far.
  sub _getNewUID {
    my $this = shift;
    my $uidRegister = TWiki::Func::getDataDir() . "/atUidReg";
    my $lockFile = "$uidRegister.lock";

    # Could do this using flock but it's not guaranteed to be
    # implemented on all systems. This technique is simpler
    # and mostly works.
    # COVERAGE OFF lock file wait
    while ( -f $lockFile ) {
      sleep(1);
    }
    # COVERAGE ON

    open( FH, ">$lockFile" ) or die "Locking $lockFile: $!";
    print FH "locked\n";
    close( FH );

    my $lastUID = 0;
    if ( -f $uidRegister ) {
      open( FH, "<$uidRegister" ) or die "Reading $uidRegister: $!";
      $lastUID = <FH>;
      close( FH );
    }

    my $uid = $lastUID + 1;
    open( FH, ">$uidRegister" ) or die "Writing $uidRegister: $!";
    print FH "$uid\n";
    close( FH );
    unlink( $lockFile ) or die "Unlocking $lockFile: $!";

    return sprintf( "%06d", $uid );
  }

  # PUBLIC when a topic containing an action is about to be saved,
  # populate these fields for the action.
  # Note: This will put a wrong date on closed actions if they were
  # closed a long while ago, but that's life.
  sub populateMissingFields {
    my $this = shift;
    my $me = _canonicalName( TWiki::Func::getWikiName() );

    if ( !defined( $this->{uid} )) {
      $this->{uid} = $this->_getNewUID();
    }

    if ( !defined( $this->{who} )) {
      $this->{who} = $me;
    }

    if ( !defined( $this->{due} )) {
      $this->{due} = $now;
    }

    if ( !defined( $this->{creator} )) {
      $this->{creator} = $me;
    }

    if ( !defined( $this->{created} )) {
      $this->{created} = $now;
    }

    if ( $this->{state} eq "closed" ) {
      if ( !defined( $this->{closer} )) {
	$this->{closer} = $me;
      }
      if ( !defined( $this->{closed} )) {
	$this->{closed} = $now;
      }
    }
  }

  # PUBLIC format as an action
  sub toString {
    my $this = shift;
    my $attrs = "";
    my $descr = "";
    foreach my $key ( keys %$this ) {
      my $type = $types{$key};
      if ( $key eq 'text') {
	$descr = $this->{text};
	$descr =~ s/^\s*(.*)\s*$/$1/os;
      } elsif ( defined( $type )) {
	if ( $type->{type} eq 'date' ) {
	  $attrs .= " $key=\"" .
	    formatTime( $this->{$key}, "attr" ) . "\"";
	} elsif ( $type->{type} !~ /noload/ ) {
	  # select or text; treat as plain text
	  $attrs .= " $key=\"" . $this->{$key} . "\"";
	}
      }
    }
    if ( $descr =~ m/<(p|br)\s*\/>/io ) {
      $descr =~ s/<p\s*\/>/\n\n/gio;
      $descr =~ s/<br\s*\/>/\n/gio;
      my $term = "EOF";
      while ( $descr =~ m/^$term/m ) {
	$term .= "F";
      }
      $descr = "<<$term\n$descr\n$term";
    }
    return "%ACTION{$attrs }% $descr";
  }

  # PRIVATE STATIC make a canonical name (including the web) for a user
  # unless it's an email address.
  sub _canonicalName {
    my $who = shift;

    return undef unless ( defined( $who ));

    if ( $who !~ /([A-Za-z0-9\.\+\-\_]+\@[A-Za-z0-9\.\-]+)/ ) {
      if ( $who eq "me" ) {
	$who = TWiki::Func::getWikiName();
      }
      if ( $who !~ /\./o ) {
	$who = "$mainweb.$who";
      }
    }
    return $who;
  }

  # PUBLIC For testing only, force current time to a known value
  sub forceTime {
    my $tim = shift;
    $now = Time::ParseDate::parsedate( $tim );
  }

  # PRIVATE get the anchor of this action
  sub getAnchor {
    my $this = shift;

    my $anchor = $this->{uid};
    if ( !$anchor ) {
      # required for old actions without uids
      $anchor = "AcTion" . $this->{ACTION_NUMBER};
    }

    return $anchor;
  }

  # PRIVATE STATIC format a time string
  sub formatTime {
    my ( $time, $format ) = @_;
    my $stime;
    if ( defined $time ) {
      if ( $format eq "attr" ) {
        $stime = localtime( $time );
	$stime =~ s/(\w+)\s+(\w+)\s+(\w+)\s+([^\s]+)\s+(\w+).*/$3-$2-$5/o;
#      } elsif ( $format eq "uid" ) {
#	my @els = localtime( $time );
#	# year yearday hour min sec
#	$stime = "$els[5]$els[7]$els[2]$els[1]$els[0]";
      } else {
        $stime = localtime( $time );
	$stime =~ s/(\w+)\s+(\w+)\s+(\w+)\s+([^\s]+)\s+(\w+).*/$1, $3 $2 $5/o;
      }
    } else {
      $stime = "BAD DATE FORMAT see Plugins.ActionTrackerPlugin#DateFormats";
    }
    return $stime;
  }

  # PUBLIC return number of seconds to go before due date, negative if action
  # is late
  sub secsToGo {
    my $this = shift;

    if ( defined( $this->{due} )) {
      return $this->{due} - $now;
    }
    return -1;
  }

  # PUBLIC return number of days to go before due date, negative if action
  # is late, 0 if it's due today
  sub daysToGo {
    my $this = shift;
    my $delta = $this->secsToGo();
    # if less that 24h late, make it a day late
    if ( $delta < 0 && $delta > -(60 * 60 * 24 )) {
      return -1;
    } else {
      return $delta / (60 * 60 * 24);
    }
  }

  # PUBLIC true if due time is before now and not closed
  sub isLate {
    my $this = shift;
    if ( $this->{state} eq "closed" ) {
      return 0;
    }
    return ( ($this->{due} - $now) < 0 );
  }

  # PRIVATE match the passed names against the given names type field.
  # The match passes if any of the names passed matches any of the
  # names in the field.
  sub _matchType_names {
    my ( $this, $vbl, $val ) = @_;

    return 0 unless ( defined( $this->{$vbl} ));

    foreach my $name ( split( /\s*,\s*/, $val )) {
      my $who = _canonicalName( $name );
      $who =~ s/\./\\./go;
      if ( $this->{$vbl} =~ m/$who,/ || $this->{$vbl} =~ m/$who$/) {
	return 1;
      }
    }
    return 0;
  }

  sub _matchType_date {
    my ( $this, $vbl, $val ) = @_;
    my $tim = Time::ParseDate::parsedate( $val, %pdopt );
    return ( defined( $this->{$vbl} ) && $this->{$vbl} == $tim );
  }

  # PRIVATE match if there are at least $val days to go before
  # action falls due
  sub _matchField_within {
    my ( $this, $val ) = @_;
    return ( $this->secsToGo() <= $val * 60 * 60 * 24 );
  }

  # PRIVATE match boolean attribute "closed"
  sub _matchField_closed {
    my $this = shift;
    return ( $this->{state} eq "closed" );
  }

  # PRIVATE match boolean attribute "open"
  sub _matchField_open {
    my $this = shift;
    return ( $this->{state} ne "closed" );
  }

  # PRIVATE match boolean attribute "late"
  sub _matchField_late {
    my $this = shift;
    return ( $this->secsToGo() < 0 && $this->{state} ne "closed" ) ? 1 : 0;
  }

  # PRIVATE trap for state so we can put state="late" into attributes
  sub _matchField_state {
    my ( $this, $val ) = @_;
    if ( $val eq "late" ) {
      return ( $this->secsToGo() < 0 && $this->{state} ne "closed" ) ? 1 : 0;
    } else {
      return ( $this->{state} eq $val );
    }
  }

  # PUBLIC true if the action matches the search attributes
  # The match is made either by calling a match function for the attribute
  # or by comparing the value of the field with the value of the
  # corresponding attribute, which is considered to be an RE.
  # To match, an action must match all conditions.
  sub matches {
    my ( $this, $a ) = @_;
    foreach my $attrName ( keys %$a ) {
      my $attrVal = $a->get( $attrName );
      my $attrType = getBaseType( $attrName );
      my $class = ref( $this );
      if ( defined( &{$class."::_matchField_$attrName"} ) ) {
	# function match
	my $fn = "_matchField_$attrName";
	if ( !$this->$fn( $attrVal )) {
	  return 0;
	}
      } elsif ( defined( $attrType ) &&
		defined( &{$class."::_matchType_$attrType"} ) ) {
	my $fn = "_matchType_$attrType";
	if ( !$this->$fn( $attrName, $attrVal )) {
	  return 0;
	}
      } elsif ( defined( $attrVal ) &&
		defined( $this->{$attrName} ) ) {
	# re match
	if ( $this->{$attrName} !~ m/$attrVal/ ) {
	  return 0;
	}
      } else {
	return 0;
      }
    }
    return 1;
  }

  # PRIVATE format the given time type
  sub _formatType_date {
    my $this = shift;
    my $fld = shift;
    return ( formatTime( $this->{$fld}, "string" ), 0 );
  }

  # PRIVATE format the given field (takes precedence of standard
  # date formatting)
  sub _formatField_due {
    my ( $this, $asHTML ) = @_;
    my $bgcol = 0;
    my $text = formatTime( $this->{due}, "string" );

    if ( !defined($this->{due}) ) {
      $bgcol = $ActionTrackerPlugin::Format::badcol;
    } elsif ( $this->isLate() ) {
      $bgcol = $ActionTrackerPlugin::Format::latecol;
      if ( !$asHTML ) {
	$text .= " (LATE)";
      }
    }

    return ( $text, $bgcol );
  }

  # PRIVATE format text field
  sub _formatField_text {
    my ( $this, $asHTML, $type ) = @_;
    my $text = $this->{text};

    if ( $asHTML && defined( $type ) && $type eq "href" ) {
      # Generate a jump-to in wiki syntax
      $text =~ s/<br ?\/?>/\n/sgo;
      # Would be nice to do the goto as a button image....
      my $jump = " (<a href=\"" .
	TWiki::Func::getViewUrl( $this->{web}, $this->{topic} ) .
	  "#" . $this->getAnchor() . "\">go to action</a>)";
      $text .= $jump;
    }
    return ( $text, 0 );
  }

  # PRIVATE format edit field
  sub _formatField_edit {
    my ( $this, $asHTML, $type, $newWindow ) = @_;

    if ( !$asHTML ) {
      # Can't edit from plain text
      return "";
    }

    my $url = "%SCRIPTURLPATH%/edit%SCRIPTSUFFIX%/" .
      $this->{web} . "/" . $this->{topic} .
	"?skin=action&action=" . $this->getAnchor();
    my $text = "<a href=\"$url\"";
    if ( $newWindow ) {
      # Javascript window call
      $text .= " onClick=\"return editWindow('$url')\"";
    }
    return ( "${text}>edit</a>", 0 );
  }

  # PUBLIC see if this other action matches according to fuzzy
  # rules. Return a number indicating the quality of the match, which
  # is the sum of:
  # action number identical - 3
  # who identical - 2
  # notify identical - 2
  # due identical - 1
  # state identical - 1
  # text identical - length of matching text
  # text sounds match - number of matching sounds
  # This is deprecated but is retained for support of non-UID actions.
  sub fuzzyMatches {
    my ( $this, $old ) = @_;
    my $sum = 0;

    # COVERAGE OFF fuzzy match with uid
    if ( defined( $this->{uid} )) {
      if ( defined( $old->{uid} ) && $this->{uid} eq $old->{uid} ) {
	return 100;
      }
      return 0;
    }

    # identical text
    if ( $this->{text} =~ m/^\Q$old->{text}\E/ ) {
      $sum += length( $this->{text} );
    } else {
      $sum += _partialMatch( $old->{text}, $this->{text} ) * 4;
    }
    if ( $this->{ACTION_NUMBER} == $old->{ACTION_NUMBER} ) {
      $sum += 3;
    }
    if ( defined( $this->{notify} ) && defined( $old->{notify} ) &&
	 $this->{notify} eq $old->{notify} ) {
      $sum += 2;
    }
    if ( defined( $this->{who} ) &&
	 $this->{who} eq $old->{who} ) {
      $sum += 2;
    }
    if ( $this->{due} == $old->{due} ) {
      $sum += 1;
    }
    if ( $this->{state} eq $old->{state} ) {
      $sum += 1;
    }
    # COVERAGE ON
    return $sum;
  }

  # PRIVATE Crude algorithm for matching text. The words in the old text
  # are matched by equality or sound and the proportion of words
  # in the old text still seen in the new text is returned.
  sub _partialMatch {
    my ( $old, $new ) = @_;
    my @aold = split( /\s+/, $old );
    my @anew = split( /\s+/, $new );
    my $matches = 0;
    foreach my $s ( @aold ) {
      for (my $t = 0; $t <= $#anew; $t++) {
	if ( $anew[$t] =~ m/^\Q$s\E$/i) {
	  $anew[$t] = "";
	  $matches++;
	  last;
	} else {
	  my $so = Text::Soundex::soundex( $s ) || "";
	  my $sn = Text::Soundex::soundex( $anew[$t] ) || "";
	  if ( $so eq $sn ) {
	    $anew[$t] = "";
	    $matches += 0.75;
	  }
	}
      }
    }
    return $matches / ( $#aold + 1 );
  }

  # PUBLIC find and format differences between this action and another
  # action, adding the changes to a hash keyed on the names of
  # people interested in notification.
  sub findChanges {
    my ( $this, $old, $format, $notifications ) = @_;

    # COVERAGE OFF safety net
    if ( !defined( $this->{notify} ) || $this->{notify} !~ m/\w/o ) {
      # Shouldn't ever get here
      return 0;
    }
    # COVERAGE ON

    my $changes = $format->formatChangesAsString( $old, $this );
    if ( $changes eq "" ) {
      return 0;
    }

    my $plain_text = $format->formatStringTable( [ $this ] );
    $plain_text .= "\n$changes\n";
    my $html_text = $format->formatHTMLTable( [ $this ], "href", 0 );
    $html_text .= $format->formatChangesAsHTML( $old, $this );

    # Add text to people interested in notification
    # in the hash
    my @notables = split(/[,\s]+/, $this->{notify} );
    foreach my $notable ( @notables ) {
      $notable = _canonicalName( $notable );
      $notifications->{$notable}{html} .= $html_text;
      $notifications->{$notable}{text} .= $plain_text;
    }

    return 1;
  }

  # PUBLIC STATIC find the action in the text with the given uid,
  # splitting the rest of the text into text before the action,
  # and text after the action.
  sub findActionByUID {
    my ( $web, $topic, $text, $uid ) = @_;

    my $sn = -1;
    if ( $uid =~ m/^AcTion(\d+)$/o ) {
      $sn = $1;
    }

    my $action;
    my $pretext = "";
    my $posttext = "";
    my $found = 0;
    my $an = 0;
    my $gathering;
    my $attrs;
    my $descr;
    my $processAction = 0;

    # FORMAT DEPENDANT ACTION SCAN
    foreach my $line ( split( /\r?\n/, $text ) ) {
      if ( $found ) {
	$posttext .= "$line\n";
      } elsif ( $gathering ) {
	if ( $line =~ m/^$gathering\b.*/ ) {
	  #$gathering = undef; PetricFrank 1 Jul 2003
	  $processAction = 1;
	} else {
	  $descr .= "$line\n";
	  next;
	}
      } elsif ( $line =~ m/^(.*?)%ACTION{(.*?)}%(.*)$/o ) {
	$pretext .= $1;
	$attrs = $2;
	$descr = $3;
	if ( $descr =~ m/\s*<<(\w+)\s*(.*)$/o ) {
	  $descr = $2;
	  $gathering = $1;
	  next;
	}
	$processAction = 1;
      } else {
	$pretext .= "$line\n";
      }

      if ( $processAction ) {
	my $anAction = new ActionTrackerPlugin::Action( $web, $topic,
							$an, $attrs, $descr );
	my $auid = $anAction->{uid};
	if ( ( defined( $auid ) && $auid eq $uid ) || $an == $sn ) {
	  $found = 1;
	  $action = $anAction;
	} else {
	  # >>> Fix from PetricFrank 1 Jul 2003
	  $pretext .= "%ACTION{$attrs}%";
	  if ( $gathering ) {
	    $pretext .= " <<$gathering\n";
	  }
	  $pretext .= "$descr\n";
	  if ( $gathering ) {
	    $pretext .= "$gathering\n";
	  }
	  $gathering = undef;
	  # <<< end of fix from PetricFrank 1 Jul 2003
	}
	$an++;
	$processAction = 0;
      }
    }
    TWiki::Func::writeDebug("Action not found $uid") unless ( $action );

    return ( $action, $pretext, $posttext );
  }

  # PUBLIC STATIC create a new action filling in attributes
  # from a CGI query as used in the action edit.
  sub createFromQuery {
    my ( $web, $topic, $an, $query ) = @_;
    my $desc = $query->param( "text" ) || "No description";
    $desc =~ s/\r?\n\r?\n/ <p \/>/sgo;
    $desc =~ s/\r?\n/ <br \/>/sgo;

    # for each of the legal attribute types, see if the query
    # contains a value for that attribute. If it does, fill it
    # in.
    my $attrs = "";
    foreach my $attrname ( keys %types ) {
      my $type = $types{$attrname};
      if ( $type->{type} !~ m/noload/o ) {
	my $val = $query->param( $attrname );
	if ( defined( $val )) {
	  $attrs .= " $attrname=\"$val\"";
	}
      }
    }
    return new ActionTrackerPlugin::Action( $web, $topic, $an, $attrs, $desc );
  }

  sub formatForEdit {
    my ( $this, $format ) = @_;

    my %expanded;
    my $table = $format->formatEditableFields( $this, \%expanded );

    foreach my $attrname ( keys %types ) {
      if ( !$expanded{$attrname} ) {
	my $type = $types{$attrname};
	if ( $type->{type} !~ m/noload/ ) {
	  $table .= $format->formatHidden( $this, $attrname );
	}
      }
    }
    return $table;
  }
}

1;
